from reportlab.lib import colors
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.platypus import TableStyle
from .settings import *


def get_styles(fontFamily=DEFAULT_FONT):
    """
    get styles for the pdf, default are :
    Title: <ParagraphStyle ‘Title’>
    Heading1: <ParagraphStyle ‘Heading1’>
    Heading2: <ParagraphStyle ‘Heading2’>
    Heading3: <ParagraphStyle ‘Heading3’>
    Heading4: <ParagraphStyle ‘Heading4’>
    Heading5: <ParagraphStyle ‘Heading5’>
    Heading6: <ParagraphStyle ‘Heading6’>
    Normal: <ParagraphStyle ‘Normal’>
    Italic: <ParagraphStyle ‘Italic’>
    BodyText: <ParagraphStyle ‘BodyText’>
    unused:
    Bullet: <ParagraphStyle ‘Bullet’>
    Definition: <ParagraphStyle ‘Definition’>
    Code: <ParagraphStyle ‘Code’>
    """
    styles = getSampleStyleSheet()
    # Titles
    styles.add(ParagraphStyle(name='cTitle', parent=styles['Title'], fontName="{}-Bold".format(fontFamily), spaceAfter=SPACE_UNIT*2))
    styles.add(ParagraphStyle(name='cHeading1', parent=styles['Heading1'], fontName="{}-Bold".format(fontFamily), spaceBefore=SPACE_UNIT,
        borderRadius=5, borderWidth=1, borderPadding=4, borderColor=colors.HexColor("#668cff"), backColor=colors.HexColor("#668cff")))
    styles.add(ParagraphStyle(name='cHeading2', parent=styles['Heading2'], fontName="{}-Bold".format(fontFamily), spaceBefore=SPACE_UNIT, leftIndent=SPACE_UNIT,
        borderRadius=5, borderWidth=1, borderPadding=4, borderColor=colors.HexColor("#809fff"), backColor=colors.HexColor("#809fff")))
    styles.add(ParagraphStyle(name='cHeading3', parent=styles['Heading3'], fontName="{}-BoldOblique".format(fontFamily), spaceBefore=SPACE_UNIT, leftIndent=SPACE_UNIT*2,
        borderRadius=5, borderWidth=1, borderPadding=4, borderColor=colors.HexColor("#99b3ff"), backColor=colors.HexColor("#99b3ff")))
    styles.add(ParagraphStyle(name='cHeading4', parent=styles['Heading4'], fontName="{}-BoldOblique".format(fontFamily), spaceBefore=SPACE_UNIT, leftIndent=SPACE_UNIT*3,
        borderRadius=5, borderWidth=1, borderPadding=4, borderColor=colors.HexColor("#b3c6ff"), backColor=colors.HexColor("#b3c6ff")))
    styles.add(ParagraphStyle(name='cHeading5', parent=styles['Heading5'], fontName="{}-Bold".format(fontFamily), spaceBefore=SPACE_UNIT, leftIndent=SPACE_UNIT*4,
        borderRadius=5, borderWidth=1, borderPadding=4, borderColor=colors.HexColor("#ccd9ff"), backColor=colors.HexColor("#ccd9ff")))
    styles.add(ParagraphStyle(name='cHeading6', parent=styles['Heading6'], fontName="{}-Bold".format(fontFamily), spaceBefore=SPACE_UNIT, leftIndent=SPACE_UNIT*5,
        borderRadius=5, borderWidth=1, borderPadding=4, borderColor=colors.HexColor("#e6ecff"), backColor=colors.HexColor("#e6ecff")))
    # Normals
    styles.add(ParagraphStyle(name='cNormal', fontName=fontFamily, parent=styles['Normal']))
    styles.add(ParagraphStyle(name='cNormal1', fontName=fontFamily, leftIndent=SPACE_UNIT, parent=styles['Normal']))
    styles.add(ParagraphStyle(name='cNormal2', fontName=fontFamily, leftIndent=SPACE_UNIT*2, parent=styles['Normal'])),
    styles.add(ParagraphStyle(name='cNormal3', fontName=fontFamily, leftIndent=SPACE_UNIT*3, parent=styles['Normal'])),
    styles.add(ParagraphStyle(name='cNormal4', fontName=fontFamily, leftIndent=SPACE_UNIT*4, parent=styles['Normal'])),
    styles.add(ParagraphStyle(name='cNormal5', fontName=fontFamily, leftIndent=SPACE_UNIT*5, parent=styles['Normal'])),
    styles.add(ParagraphStyle(name='cNormal6', fontName=fontFamily, leftIndent=SPACE_UNIT*6, parent=styles['Normal'])),
    # BodyText
    styles.add(ParagraphStyle(name='cBodyText', fontName=fontFamily, parent=styles['BodyText']))
    # Italic
    styles.add(ParagraphStyle(name='cItalic', fontName="{}-Oblique".format(fontFamily), parent=styles['Italic']))

    return styles


def get_title_style(styles, level):
    """ get title style by level """
    if level == 0:
        return styles["cTitle"]
    elif level in [1, 2, 3, 4, 5, 6]:
        return styles["{}{}".format("cHeading", str(level))]
    else:
        return styles["cItalic"]


def get_normal_style(styles, level):
    """ get title style by level """
    if level == 0:
        return styles["cNormal"]
    elif level in [1, 2, 3, 4, 5, 6]:
        return styles["{}{}".format("cNormal", str(level))]
    else:
        return styles["cNormal6"]


def get_table_style():
    """ get table style """
    return TableStyle([
        ('BACKGROUND',(0,0),(-1,0),colors.HexColor("#ccd9ff")),
        ('ALIGN', (1,1), (-1,-1), 'RIGHT'),
        ('INNERGRID', (0,0), (-1,-1), 0.25, colors.HexColor("#ccd9ff")),
        ('BOX', (0,0), (-1,-1), 2, "#99b3ff")
    ], hAlign='LEFT', wordWrap='LTR')



import unittest
import os.path
from xml_archive_to_pdf.main import main


"""
Main unit tests
"""


class MainTest(unittest.TestCase):

    def setUp(self):
        self.xml_file = "tests/data/pathfinder_1.xml"
        self.pdf_file = "tests/data/pathfinder_1.pdf"
        self.logo_file = "tests/data/logo.png"
        self.font_folder = "tests/data/CustomFont"
        self.args = {
            "--input": self.xml_file,
            "--output": self.pdf_file,
            "--logo": self.logo_file,
            "--font": self.font_folder
        }

    def test_main(self):
        if os.path.exists(self.pdf_file):
            os.remove(self.pdf_file)
        main(args=self.args)
        self.assertTrue(os.path.exists(self.pdf_file))
        self.assertTrue(os.path.getsize(self.pdf_file) > 0)



"""
All utils
"""
import xml.etree.cElementTree as ET
import re
from reportlab.lib.pagesizes import letter
from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, Image, Table, Indenter
from .styles import get_styles, get_title_style, get_normal_style, get_table_style
from .settings import *
from reportlab.lib.units import mm
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from os.path import join


def get_bare_tag(elem):
    """ Get the tag name without namespace """
    return elem.tag.rsplit('}', 1)[-1]


def get_clean_text(elem):
    """ return value without line breaker chars """
    return re.sub(r"[\f\n\r\t\v]", "", elem.text).strip() if elem.text else ""


def get_doc(pdf_file):
    """ get the pdf doc """
    return SimpleDocTemplate(pdf_file, pagesize=letter, rightMargin=SPACE_UNIT*2,
                             leftMargin=SPACE_UNIT*2, topMargin=SPACE_UNIT*2, bottomMargin=SPACE_UNIT*2)


def registerFont(fontFamilyName, folder_path):
    """ register a font """
    # Register a custom font
    pdfmetrics.registerFont(TTFont('{}-Bold'.format(fontFamilyName), join(folder_path, "{}-Bold.ttf".format(fontFamilyName))))
    pdfmetrics.registerFont(TTFont(fontFamilyName, join(folder_path, "{}.ttf".format(fontFamilyName))))
    pdfmetrics.registerFont(TTFont('{}-Oblique'.format(fontFamilyName), join(folder_path, "{}-Oblique.ttf".format(fontFamilyName))))
    pdfmetrics.registerFont(TTFont('{}-BoldOblique'.format(fontFamilyName), join(folder_path, "{}-BoldOblique.ttf".format(fontFamilyName))))
    pdfmetrics.registerFontFamily(
        fontFamilyName,
        normal=fontFamilyName,
        bold='{}-Bold'.format(fontFamilyName),
        italic='{}-Italic'.format(fontFamilyName),
        boldItalic='{}-BoldItalic'.format(fontFamilyName)
    )


def get_label(e):
    """ get a label or a title from element """
    return e.attrib.get(ATTR_NAME) if e.attrib.get(ATTR_NAME) is not None else get_bare_tag(e)


def has_children(e):
    """ true if element has children """
    return True if e.getchildren() else False


def add_logo(Story, logo_file):
    """ add a logo """
    im = Image(logo_file)
    im.hAlign = 'RIGHT'
    Story.append(im)
    Story.append(Spacer(1, SPACE_UNIT))


def ptTomm(pt):
    """ convert pt to mm """
    return pt * 0.352778


def write_table(Story, e, level, styles):
    """ write a full table """
    data = []
    colWidths = []
    # On récupère les lignes
    rows = e.getchildren()
    label = get_label(e)
    # Si on a un label non vide, on l'affiche comme titre
    if label:
        Story.append(Paragraph(label, get_title_style(styles, level)))
    Story.append(Spacer(1, SPACE_UNIT))
    # On affiche le table
    if rows:
        # On récupère les titres du tableau à partir du 1er element
        columns_names = list(map(lambda x: Paragraph(get_label(x), styles['cBodyText']), rows[0].getchildren()))
        data.append(columns_names)
        colWidths = [((ptTomm(letter[0])-(SPACE_UNIT*2))/len(columns_names))*mm for i in range(0, len(columns_names))]
        # On parcourt les éléments pour les insérer dans le tableau
        for row in rows:
            values = row.getchildren()
            if not values or has_children(values[0]):
                raise Exception("Table {} has malformed elements".format(label))
            if len(columns_names) != len(values):
                raise Exception("Table {} has wrong elements".format(label))
            line_values = list(map(lambda x: Paragraph(get_clean_text(x), styles['cBodyText']), values))
            data.append(line_values)

        # Build the table
        t = Table(data, colWidths=colWidths)
        #t.setStyle(get_table_style(level))
        t.setStyle(get_table_style())
        # Story.append(Indenter(left=(level*SPACE_UNIT)))
        Story.append(t)
        # Story.append(Indenter(left=-level*SPACE_UNIT))
    return Story


def write_elem(Story, e, level, styles):
    """ write element in the pdf """
    # get label and value
    label = get_label(e)
    value = get_clean_text(e)
    # Gestion des titres: un titre est un élément avec des enfants
    if has_children(e):
        # Si on a un label non vide, on l'affiche
        if label:
            Story.append(Paragraph(label, get_title_style(styles, level)))
        # Si on a forcé le nom du label à vide, on met juste un petit espace
        else:
            Story.append(Spacer(1, SPACE_UNIT))
    # C'est un element de type clé-valeur, on l'affiche normalement
    else:
        if label or value:
            Story.append(Paragraph("<b>{} :</b> {}".format(label, value), get_normal_style(styles, level)))
    return Story


def is_writable_element(e, action, table_level):
    """ return true if the element is writable """
    return ET.iselement(e) and action == EVENT_START and not table_level[0]


def calcul_level(action, level):
    """ calcul the level """
    if (action == EVENT_START):
        level += 1
    elif (action == EVENT_END):
        level -= 1
    return level


def is_leaving_table(action, level, table_level):
    """ check if the element/action is leaving a table """
    return action == EVENT_END and table_level[0] and table_level[1] == level


def build_pdf(xml_file, pdf_file, logo_file=None, font_folder=None):
    """ build the pdf file """
    # Build the doc and the story
    doc = get_doc(pdf_file)
    # register a font and get styles
    if font_folder:
        registerFont("CustomFont", font_folder)
        styles = get_styles("CustomFont")
    else:
        styles = get_styles()
    Story = []
    # events and level for lxml
    level = 0
    # Add logo
    if logo_file:
        add_logo(Story, logo_file)
    # Variable to know if the element is a table
    table_level = (False, 0)
    # Parse the full xml
    for action, e in ET.iterparse(xml_file, events=(EVENT_START, EVENT_END)):
        # If a new element is encountered
        # Ignore elements in a table (because the table is builded entirely before)
        if is_writable_element(e, action, table_level):
            # For style table
            attr_style = e.attrib.get(ATTR_STYLE)
            # Build a full table
            if attr_style and attr_style == ATTR_STYLE_TABLE:
                table_level = (True, level)
                Story = write_table(Story, e, level, styles)
            # For title and simple key value
            else:
                Story = write_elem(Story, e, level, styles)
        # Calcul the level
        level = calcul_level(action, level)
        # Leave a table
        if is_leaving_table(action, level, table_level):
            table_level = (False, 0)
    # build
    doc.build(Story)
